/***************************************************************************
                          options.h  -  description
                             -------------------

    functionality        : defines many options and the option class
    relevant classes     : options
    relevant functions   : default_options, hardwired_options, trau_opcions

    begin                : Fri Oct 26 2001
    copyright            : (C) 2001 by C.Ferri & J. Hernndez-Orallo
    email                : {cferri, jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef OPTIONS_H
#define OPTIONS_H

#include <string>
#include <vector>
#include "criteria.h"
#include "evaluate.h"
#include "components.h"


// Sets options according to some predefined defaults
void default_options(options &Options);

// Sets options according to some predefined defaults
void hardwired_options(options &Options);

// Sets options according to the usual UNIX command options.
void trau_opcions( char *const *argv,const int argc, options &Options, string &ftrain, string & ftest, string &fcost, string &ftestcost, const string &defaultdataset, const string &);


/////////////////////////  INPUT OPTIONS

typedef enum missing_numeric_values { IGNORE_EXAMPLES_WITH_MISSING_NUMERIC_VALUES, SUBSTITUTE_MISSING_NUMERIC_VALUES };

/////////////////////////  INPUT OPTIONS

typedef enum  show_mumber_solutions { DO_NOT_SHOW_NUMBER_SOLUTIONS,SHOW_NUMBER_SOLUTIONS};



/////////////////////////  OUTPUT OPTIONS

typedef enum show_rules_mode { FUNCLOGIC, IF_THEN_ELSE, FUNCLOGIC_SIMPLIFIED };
typedef enum show_class_distribution { YES, NO };
typedef enum show_all_multitree_rules { NO_SHOW_ALL_RULES, SHOW_ALL_RULES, ALL_FILES_TO_FILE };
typedef enum show_all_kbest_solutions { NO_SHOW_SOLUTIONS, SHOW_SOLUTIONS, ALL_SOLUTIONS_TO_FILE };
typedef enum show_solutions_components { NO_SHOW_COMPONENTS, SHOW_COMPONENTS };
typedef enum show_statistics { JUST_ACCURACY, ABSOLUTE_STATISTICS, RELATIVE_STATISTICS, BOTH_STATISTICS };
typedef enum show_confusion_matrix { NO_SHOW_CONFUSION_MATRIX, SHOW_CONFUSION_MATRIX, SHOW_ONLY_IF_COSTS};
typedef enum compute_ROC_points { NO_COMPUTE_ROC_POINTS, COMPUTE_ROC_POINTS };
typedef enum show_ROC_points { NO_SHOW_ROC_POINTS, SHOW_ROC_POINTS };
typedef enum compute_ROC_area { NO_COMPUTE_ROC_AREA, COMPUTE_ROC_AREA_BY_HAND  };
typedef enum generate_ROC_curve_file { NO_GENERATE_ROC_CURVE_FILE, GENERATE_ROC_CURVE_FILE };

struct output_options {
  show_rules_mode ShowRulesMode;                // not used
  show_class_distribution ShowClassDist;        // not used
  show_all_multitree_rules ShowAllMultiTreeRules;
  show_all_kbest_solutions ShowAllKBestSolutions;
  show_solutions_components ShowSolutionsComponents;
  show_confusion_matrix ShowConfusionMatrix;
  show_statistics ShowStatistics;
  compute_ROC_points ComputeROCPoints;
  show_ROC_points ShowROCPoints;
  compute_ROC_area ComputeROCArea;
  generate_ROC_curve_file GenerateROCCurveFile;
};



/////////////////////////  EXPECTEDERROR OPTIONS

typedef enum expected_error_method { NO_COMPUTE,
                                                              RELATIVE_FREQUENCY_WITH_MAJORITY_CLASS,
                                                              RELATIVE_FREQUENCY_WITH_FREQ_PROB,
                                                              COST_WITH_MINIMUM_CLASS,
                                                              COST_WITH_FREQ_PROB,
                                                              COST_WITH_COST_PROB };


struct expected_error_options {
  expected_error_method Method;
};




/////////////////////////  FREQUENCY ERROR SMOOTHING (used for class probabilities and node probabilities)

typedef enum smoothing_method { NO_SMOOTHING, LAPLACE, K_ESTIMATE, M_ESTIMATE, M_ESTIMATE_UNIFORM };

struct smoothing_options {
  smoothing_method Method;
  int m;
};



/////////////////////////  MULTITREE OPTIONS

typedef enum multitree_construction { ONE_ONLY, ONE_AND_FORGET, MANY_AND_FORGET, MANY_AND_MAINTAIN, THRUSTING };

typedef enum multitree_suspended_nodes_forgetting { MAINTAIN_ALL, MAINTAIN_LOG_RANDOM,
                                                    MAINTAIN_CONST_RANDOM, MAINTAIN_LOG_RANDOM_WITH_DEPTH,
                                                    MAINTAIN_LOG_RANDOM_WITH_SQUARED_DEPTH,
                                                    MAINTAIN_LOG_RANDOM_WITH_DEPTH_ADJUSTED };

struct multitree_options {
  multitree_construction Construction;
  best_tree_criterion BestTreeCrit;
  second_tree_criterion SecondTreeCrit;
  long MaxNumTrees;
  multitree_suspended_nodes_forgetting SuspendedNodesForgetting;
  long SuspendedNodesConstMaintain;
};



/////////////////////////  COMBINATION OPTIONS

typedef enum solution_combination{NO_COMBINATION,CROSS_COVERAGE_COMBINATION,MAJORITY_CRISP,MAJORITY_ABSOLUT_STOCHASTIC,
MAJORITY_RELATIVE_STOCHASTIC,MAJORITY_COST_STOCHASTIC};
typedef enum combination_vector { RELATIVE, ABSOLUTE };
typedef enum combination_vector_method { ORIGINAL, GOOD_LOSER, BAD_LOSER, DIFFERENCE, MAJORITY, SQUARED };
typedef enum fusion_method { SUM, PROD, ARITHMEAN, GEOMEAN, MAXIMUM, MINIMUM, MEDIAN };

struct combination_options {
  solution_combination Comb;
  bool Smoothing;
  combination_vector CombinationVector;
  combination_vector_method CombinationVectorMethod;
  fusion_method FusionMethod;
  bool PruneInCombination;
};


/////////////////////////  TEST COST OPTIONS

enum test_cost_method { NO_TEST_COSTS, UNIFORM_TEST_COSTS, TEST_COSTS_FROM_FILE };

enum test_cost_use { TEST_COSTS_NO_USE, TEST_COSTS_LINEAR_PLUS1_WITHOUT_REPETITION, TEST_COSTS_LINEAR_PLUS1_WITH_REPETITION };





/////////////////////////  WEIGHTS OPTIONS (MISCLASSIFICATION COSTS)

enum weights_method { NO_COSTS,
                      UNIFORM_WEIGHTS,  // Sets a matrix:
                                                                           //         (1, 1, 1)
                                                                           //         (1, 1, 1)
                                                                           //         (1, 1, 1)
                                                                          // Equivalent to NO_WEIGHTS
                      WEIGHTS_FROM_FILE,  // Given the weights (a, b, c), sets a matrix:
                                     										// (a, -b, -c)
                                     										// (-a, b, -c)
                                     										 // (-a, -b, c)
                                     										 // This is equivalent to oversampling
                      INVERSE_FREQUENCY_WEIGHTS,  // Sets the weights as the inverse of the frequency of each class from the dataset and then constructs the matrix as the previous opiont
                      COSTS_FROM_MATRIX };


enum class_selection_method { MAJORITY_CLASS,
                              MIN_COST_CLASS,
                              MIN_COST_CLASS_WITHOUT_SMOOTHING,
                              STRATIFICATION_CLASS,
                              STRATIFICATION_CLASS_WITHOUT_SMOOTHING };


struct weights_options {
  weights_method Method;
  class_selection_method ClassSelection;
  float Parameter;
};


/////////////////////////  USE COSTS FOR COMPUTING A PROBABILITY

enum cost_derived_probability_method { DIRECT, WITH_SMOOTHING };

/////////////////////////  USE COSTS FOR COMPUTING THE PROBABILITY FOR SPLITS

enum probability_in_splitting_criteria { FROM_FREQUENCY_NO_SMOOTHING,
                                                           FROM_FREQUENCY_SMOOTHING, FROM_COSTS,
                                                            FROM_FREQUENCY_WITH_STRATIFICATION,
                                                            FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING
                                                             };


///////////////                     COMPONENTS IN SPLIT CRITERION
enum components_in_split_criterion_method { NO_USE_OF_COMPONENT_IN_SPLIT, ADAPTATIVE_COMPONENT,
                                           RANDOM_COMPONENTS } ;


class components_in_split_criterion {
public:
  components_in_split_criterion_method Met;
 float Factor;

public:
  components_in_split_criterion(){
  }

  components_in_split_criterion(components_in_split_criterion_method Mt, float i=0) {
    Met=Mt;
    Factor=i;
  }
};


////////////////////////////   K-BEST SELECTION METHOD

enum  k_best_selection_method { K_BEST_LESS_VISITED, K_BEST_LESS_VISITED_PLUS,
                                K_BEST_DIFFERENT_COMPONENTS,
                                K_BEST_LESS_VISITED_THEN_DIFFERENT_COMPONENTS,
                                K_BEST_RANDOM };


////////////////////////////   K-BEST SELECTION METHOD USE OF COMPONENTS

enum k_best_selection_component_generation { COMP_ACCUMULATE, COMP_RANDOM_GENERATE_FROM_START,
                                             COMP_RANDOM_GENERATE_FROM_SECOND };


class k_best_selection {
public:
  k_best_selection_method  Met;
  long NumSolutions;

  float FactorForRandomComponents;
  k_best_selection_component_generation ComponentsGeneration;

public:
  k_best_selection(){
  }

  k_best_selection(k_best_selection_method Mt, float i=0) {
    Met=Mt;
//    Factor=i;
  }
};


////////////////////////////   COMB2SINGLE METHODS

enum comb2single_method { INVENTED_DATASET, NO_EXTRACTION };

enum similarity_method { KAPPA, KAPPA1, QSTAT };

class comb2single_options {
public:
  comb2single_method Method;
  long InventedDataSetLength;
  similarity_method SimilarityMethod;

  double SimilarityImportanceFactor;
  double OccamImportanceFactor;
  double TestCostImportanceFactor;
  bool Use_Occam;
  bool Use_TestCost;
};


//////////////////////////// SAMPLING AND CROSS-VALIDATION OPTIONS


enum  sample_training_set { NO_SAMPLE_TRAINING_SET, SAMPLE_TRAINING_SET };

enum  cross_validation { USE_SEPARATE_TEST_SET, CROSS_VALIDATION, KFOLD_CROSS_VALIDATION, REPEATED_KFOLD_CROSS_VALIDATION };



/////////////////////////// RELIABILITY CALCULATION USED IN FUNCTIONS THAT PREDICT

enum reliability_calculation { FREQUENCY_RELIABILITY, LAPLACE_RELIABILITY };

/////////////////////////// CLASS PROBABILITY VECTOR CALCULATION USED IN FUNCTIONS THAT PREDICT (AND ULTIMATELY USED FOR COMPUTING AUC EXAMPLE BY EXAMPLE)

enum class_probability_vector_calculation { FREQUENCY_CLASS_PROBABILITY_VECTOR, LAPLACE_CLASS_PROBABILITY_VECTOR };

/////////////////////////  STRUCT OPTIONS

struct options {

  missing_numeric_values MissingNumericValues;
  reliability_calculation ReliabilityCalculation;
  class_probability_vector_calculation ClassProbabilityVectorCalculation;
  show_mumber_solutions ShowNumberSolutions;
  expected_error_options ExpectedError;
  smoothing_options Smoothing;
  smoothing_options NodeSmoothing;
  probability_in_splitting_criteria ProbabilityInSplittingCriteria;
  output_options Output;
  multitree_options Multitree;
  combination_options Comb;
  split_selection_criteria SSC;
  pruning_criterion PC;
  post_pruning_method PostPruningMethod;
  numerical_interval_criterion IntervalCriterion;
  vector<bool> Enabled_Partitions;

  weights_options Weights;
  cost_derived_probability_method Cost_Derived_Probability_Method;
  class_matrix CostMatrix;
  vector<double> WeightVector;  // Cuidao perque s'ha d'inicialitzar a m. Ha d'estar congruent sempre amb la CostMatrix

  vector<double> TestCostVector;  // vector of attribute test costs.
  test_cost_method TestCostMethod;
  test_cost_use TestCostUse;
  double TestCostRelevanceInSplitting; // must be from 0 to infinite.
  double TestCostRelevanceInSelectBest; // must be from 0 to 1.

  component_matrix Components;
  component_matrix AccComponents;
  component_matrix RandomComponents;
  bool Components_Reckoning;  // if this is true the previous matrices are computed
  components_in_split_criterion Components_In_Split_Criterion;
  float ComponentsRandomFactor; // value between 0 and 1 for generating random component matrices

  k_best_selection SelectKBest;

  comb2single_options Comb2Single;

  sample_training_set SampleTrainingSet;
  cross_validation CrossValidation;
  float SampleTrainingSetProportion;
  int kFoldValidation;
  int RepeatKFold;
  bool AllowTestWithoutOneClass;

  float SecondTreeDepthAccStat;  // Statistic about the depth where the second tree is opened. It's not an option

  options() : CostMatrix(1) { // A EST FATAL PERO CostMatrix no t constructor buit
  }

  ~options() {
//    cout << "Options destroyed" << endl;
  }

  bool Validate(int NClasses) {
    if ((PC.Met == MDL_PRUNING)  &&
          ( (SSC != DESC_MDL) && (SSC != WEIGHTED_DESC_MDL)) )  {
      cout << "To be implemented: MDL_based pruning cannot be done if MDL selection is not activated. Pruning will not be used" << endl;
      PC.Met = NO_PRUNING;
    }

    if ((SSC == LOCAL_ROC_AREA) || (SSC == ONE_POINT_LOCAL_ROC_AREA)) {
      if (NClasses != 2) {
        cout << "Local ROC Area Splitting Criterion cannot be used if the Number of Classes is different than 2" << endl;
        return false;
      }
    }


    if ((Comb2Single.Method != NO_EXTRACTION) && (Comb.Comb == NO_COMBINATION)) {
      cout << "Error in Options: You can't convert from combination to single solution without using any combination option" << endl;
      // exit(-1);
      return false;
    }

   // FALTEN MS COMPROVACIONS



    if (Comb2Single.SimilarityImportanceFactor != 0.0)
      ; // do nothing

    if (Comb2Single.OccamImportanceFactor != 0.0)
      Comb2Single.Use_Occam= true;
    else
      Comb2Single.Use_Occam= false;

    if (Comb2Single.TestCostImportanceFactor != 0.0)
      Comb2Single.Use_TestCost= true;
    else
      Comb2Single.Use_TestCost= false;


    return true;
  }
};


#endif

